@@ -2,20 +2,21 @@
 # This file is part of Quark-Engine - https://github.com/quark-engine/quark-engine
 # See the file 'LICENSE' for copying permission.
 
-import click
 import json
-import numpy as np
 import os
+
+import click
+import numpy as np
 from tqdm import tqdm
 
-from quark import __version__
-from quark import config
+from quark import __version__, config
+from quark.core.parallelquark import ParallelQuark
 from quark.core.quark import Quark
 from quark.core.struct.ruleobject import RuleObject
 from quark.logo import logo
 from quark.utils.colors import yellow
-from quark.utils.graph import show_comparison_graph, select_label_menu
-from quark.utils.pprint import print_success, print_info, print_warning
+from quark.utils.graph import select_label_menu, show_comparison_graph
+from quark.utils.pprint import print_info, print_success, print_warning
 from quark.utils.weight import Weight
 
 logo()
@@ -64,8 +65,10 @@ logo()
 @click.option(
     "-g",
     "--graph",
-    is_flag=True,
+    is_flag=False,
     help="Create call graph to call_graph_image directory",
+    flag_value="png",
+    type=click.Choice(["png", "json"]),
     required=False,
 )
 @click.option(
@@ -118,6 +121,14 @@ logo()
     required=False,
     default="androguard",
 )
+@click.option(
+    "--multi-process",
+    "num_of_process",
+    type=click.IntRange(min=1),
+    help="Allow analyzing APK with N processes",
+    required=False,
+    default=1,
+)
 def entry_point(
     summary,
     detail,
@@ -132,26 +143,58 @@ def entry_point(
     label,
     comparison,
     core_library,
+    num_of_process,
 ):
     """Quark is an Obfuscation-Neglect Android Malware Scoring System"""
 
-    # Load APK
-    data = Quark(apk[0], core_library)
-
     # Load rules
-    rules_list = [file for file in os.listdir(rule) if file.endswith("json")]
+    rule_buffer_list = []
+    rule_filter = summary or detail
+
+    # Determine the location of rules
+    if rule_filter and rule_filter.endswith("json"):
+        if not os.path.isfile(rule_filter):
+            print_warning(
+                f"Specified rule not found.\n"
+                f"If you want to specify one of the rules of Quark-Rule, "
+                f"use {yellow(f'{config.DIR_PATH}/{rule_filter}')} "
+                f"as an argument."
+            )
+            return
+
+        rule_path_list = [rule_filter]
+    else:
+        rule_path_list = [
+            os.path.join(rule, file)
+            for file in os.listdir(rule)
+            if file.endswith("json")
+        ]
+
+    # Load rules into memory
+    update_rule_buffer(rule_buffer_list, rule_path_list)
+
+    # Determine if the user provide a rule label
+    if (
+        rule_filter
+        and rule_filter != "all_rules"
+        and not rule_filter.endswith("json")
+    ):
+        rule_checker_list = [
+            rule_checker
+            for rule_checker in rule_buffer_list
+            if rule_filter in rule_checker.label
+        ]
+    else:
+        rule_checker_list = rule_buffer_list
 
     if comparison:
 
         # selection of labels on which it will be done the comparison on radar chart
         # first look for all label found in the rule list
-        all_labels = set()
-        for single_rule in tqdm(rules_list):
-            rulepath = os.path.join(rule, single_rule)
-            rule_checker = RuleObject(rulepath)
-            labels = rule_checker.label  # array type, e.g. ['network', 'collection']
-            for single_label in labels:
-                all_labels.add(single_label)
+        all_labels = set()  # array type, e.g. ['network', 'collection']
+
+        for rule_checker in rule_checker_list:
+            all_labels.update(rule_checker.label)
 
         # let user choose a list of label on which it will be performed the analysis
         selected_label = np.array(
@@ -161,7 +204,11 @@ def entry_point(
         # perform label based analysis on the apk_
         malware_confidences = {}
         for apk_ in apk:
-            data = Quark(apk_)
+            data = (
+                ParallelQuark(apk_, core_library, num_of_process)
+                if num_of_process > 1
+                else Quark(apk_, core_library)
+            )
             all_labels = {}
             # dictionary containing
             # key: label
@@ -169,15 +216,17 @@ def entry_point(
             # $ print(all_rules["accessibility service"])
             # > [60, 40, 60, 40, 60, 40]
 
-            for single_rule in tqdm(rules_list):
-                rulepath = os.path.join(rule, single_rule)
-                rule_checker = RuleObject(rulepath)
+            # analyse malware only on rules where appears label selected
+            rule_checker_list = [
+                rule_checker
+                for rule_checker in rule_buffer_list
+                if len(np.intersect1d(rule_checker.label, selected_label)) != 0
+            ]
 
-                # analyse malware only on rules where appears label selected
-                labels = np.array(rule_checker.label)
-                if len(np.intersect1d(labels, selected_label)) == 0:
-                    continue
+            if num_of_process > 1:
+                data.apply_rules(rule_checker_list)
 
+            for rule_checker in tqdm(rule_checker_list):
                 # Run the checker
                 data.run(rule_checker)
                 confidence = rule_checker.check_item.count(True) * 20
@@ -197,7 +246,9 @@ def entry_point(
                 # on radar data use the maximum confidence for a certain label
                 radar_data[_label] = np.max(confidences)
 
-            radar_confidence = [value_ for _label, value_ in radar_data.items()]
+            radar_confidence = [
+                value_ for _label, value_ in radar_data.items()
+            ]
             malware_confidences[apk_.split("/")[-1]] = radar_confidence
 
         show_comparison_graph(
@@ -207,6 +258,15 @@ def entry_point(
             font_size=22,
         )
 
+        return
+
+    # Load APK
+    data = (
+        ParallelQuark(apk[0], core_library, num_of_process)
+        if num_of_process > 1
+        else Quark(apk[0], core_library)
+    )
+
     if label:
         all_labels = {}
         # dictionary containing
@@ -215,13 +275,16 @@ def entry_point(
         # $ print(all_rules["accessibility service"])
         # > [60, 40, 60, 40, 60, 40]
 
-        for single_rule in tqdm(rules_list):
-            rulepath = os.path.join(rule, single_rule)
-            rule_checker = RuleObject(rulepath)
+        if num_of_process > 1:
+            data.apply_rules(rule_buffer_list)
+
+        for rule_checker in tqdm(rule_buffer_list):
             # Run the checker
             data.run(rule_checker)
             confidence = rule_checker.check_item.count(True) * 20
-            labels = rule_checker.label  # array type, e.g. ['network', 'collection']
+            labels = (
+                rule_checker.label
+            )  # array type, e.g. ['network', 'collection']
             for single_label in labels:
                 if single_label in all_labels:
                     all_labels[single_label].append(confidence)
@@ -244,28 +307,18 @@ def entry_point(
     # Show summary report
     if summary:
 
-        if summary == "all_rules":
-            label_flag = False
-        elif summary.endswith("json"):
-            rules_list = [summary]
-            label_flag = False
-        else:
-            label_flag = True
-
-        for single_rule in tqdm(rules_list):
-            rulepath = os.path.join(rule, single_rule)
-            rule_checker = RuleObject(rulepath)
-
-            labels = rule_checker.label
-            if label_flag and summary not in labels:
-                continue
+        if isinstance(data, ParallelQuark):
+            data.apply_rules(rule_checker_list)
 
+        for rule_checker in tqdm(rule_checker_list):
             # Run the checker
             data.run(rule_checker)
 
             data.show_summary_report(rule_checker, threshold)
 
-        w = Weight(data.quark_analysis.score_sum, data.quark_analysis.weight_sum)
+        w = Weight(
+            data.quark_analysis.score_sum, data.quark_analysis.weight_sum
+        )
         print_warning(w.calculate())
         print_info(f"Total Score: {data.quark_analysis.score_sum}")
         print(data.quark_analysis.summary_report_table)
@@ -273,34 +326,29 @@ def entry_point(
         if classification:
             data.show_rule_classification()
         if graph:
-            data.show_call_graph()
+            data.show_call_graph(graph)
 
     # Show detail report
     if detail:
+        threshold_number = int(threshold) if threshold else 0
 
-        if detail == "all_rules":
-            label_flag = False
-        elif detail.endswith("json"):
-            rules_list = [detail]
-            label_flag = False
-        else:
-            label_flag = True
-
-        for single_rule in tqdm(rules_list):
-            rulepath = os.path.join(rule, single_rule)
-            rule_checker = RuleObject(rulepath)
-
-            labels = rule_checker.label
-            if label_flag and detail not in labels:
-                continue
+        if isinstance(data, ParallelQuark):
+            data.apply_rules(rule_checker_list)
 
+        for rule_checker in tqdm(rule_checker_list):
             # Run the checker
             data.run(rule_checker)
 
-            print(f"Rulepath: {rulepath}")
-            print(f"Rule crime: {rule_checker.crime}")
-            data.show_detail_report(rule_checker)
-            print_success("OK")
+            confidence = rule_checker.check_item.count(True) * 20
+
+            if confidence >= threshold_number:
+                print(
+                    f"Rulepath: "
+                    f"{os.path.join(rule, rule_checker.rule_filename)}"
+                )
+                print(f"Rule crime: {rule_checker.crime}")
+                data.show_detail_report(rule_checker)
+                print_success("OK")
 
         if classification:
             data.show_rule_classification()
@@ -310,10 +358,10 @@ def entry_point(
     # Show JSON report
     if output:
 
-        for single_rule in tqdm(rules_list):
-            rulepath = os.path.join(rule, single_rule)
-            rule_checker = RuleObject(rulepath)
+        if isinstance(data, ParallelQuark):
+            data.apply_rules(rule_buffer_list)
 
+        for rule_checker in tqdm(rule_buffer_list):
             # Run the checker
             data.run(rule_checker)
 
@@ -342,6 +390,20 @@ def entry_point(
         for p in data.apkinfo.permissions:
             print(p)
 
+    if isinstance(data, ParallelQuark):
+        data.close()
+
+
+def update_rule_buffer(rule_buffer_list, rule_path_list):
+    for rule_path in tqdm(rule_path_list):
+        with open(rule_path, "r") as json_file:
+            json_data = json.loads(json_file.read())
+        if isinstance(json_data, list):
+            for rule_data in json_data:
+                rule_buffer_list.append(RuleObject(rule_path, rule_data))
+        else:
+            rule_buffer_list.append(RuleObject(rule_path, json_data))
+
 
 if __name__ == "__main__":
     entry_point()
