@@ -2,9 +2,10 @@
 # This file is part of Quark-Engine - https://github.com/quark-engine/quark-engine
 # See the file 'LICENSE' for copying permission.
 
-import numpy as np
 import operator
 import os
+
+import numpy as np
 import pandas as pd
 
 from quark.core.analysis import QuarkAnalysis
@@ -48,7 +49,9 @@ class Quark:
         elif core_library == "androguard":
             self.apkinfo = AndroguardImp(apk)
         else:
-            raise ValueError(f"Unsupported core library for Quark: {core_library}")
+            raise ValueError(
+                f"Unsupported core library for Quark: {core_library}"
+            )
 
         self.quark_analysis = QuarkAnalysis()
 
@@ -107,7 +110,9 @@ class Quark:
                 depth, first_method_set, second_method_set
             )
 
-    def method_recursive_search(self, depth, first_method_set, second_method_set):
+    def method_recursive_search(
+        self, depth, first_method_set, second_method_set
+    ):
         # Not found same method usage, try to find the next layer.
         depth += 1
         if depth > MAX_SEARCH_LAYER:
@@ -120,14 +125,22 @@ class Quark:
         # Extend the xref from function into next layer.
         for method in first_method_set:
             if self.apkinfo.upperfunc(method):
-                next_level_set_1 = self.apkinfo.upperfunc(method) | next_level_set_1
+                next_level_set_1 = (
+                    self.apkinfo.upperfunc(method) | next_level_set_1
+                )
         for method in second_method_set:
             if self.apkinfo.upperfunc(method):
-                next_level_set_2 = self.apkinfo.upperfunc(method) | next_level_set_2
+                next_level_set_2 = (
+                    self.apkinfo.upperfunc(method) | next_level_set_2
+                )
 
-        return self.find_intersection(next_level_set_1, next_level_set_2, depth)
+        return self.find_intersection(
+            next_level_set_1, next_level_set_2, depth
+        )
 
-    def check_sequence(self, mutual_parent, first_method_list, second_method_list):
+    def check_sequence(
+        self, mutual_parent, first_method_list, second_method_list
+    ):
         """
         Check if the first function appeared before the second function.
 
@@ -155,9 +168,14 @@ class Quark:
                 # seq_table would look like: [(getLocation, 1256), (sendSms, 1566), (sendSms, 2398)]
 
                 method_list_need_check = [x[0] for x in seq_table]
-                sequence_pattern_method = [first_call_method, second_call_method]
+                sequence_pattern_method = [
+                    first_call_method,
+                    second_call_method,
+                ]
 
-                if tools.contains(sequence_pattern_method, method_list_need_check):
+                if tools.contains(
+                    sequence_pattern_method, method_list_need_check
+                ):
                     state = True
 
                     # Record the mapping between the parent function and the wrapper method
@@ -192,7 +210,9 @@ class Quark:
                 pyeval = PyEval(self.apkinfo)
                 # Check if there is an operation of the same register
 
-                for bytecode_obj in self.apkinfo.get_method_bytecode(parent_function):
+                for bytecode_obj in self.apkinfo.get_method_bytecode(
+                    parent_function
+                ):
                     # ['new-instance', 'v4', Lcom/google/progress/SMSHelper;]
                     instruction = [bytecode_obj.mnemonic]
                     if bytecode_obj.registers is not None:
@@ -220,7 +240,9 @@ class Quark:
                             ):
                                 state = True
 
-                                if keyword_item_list and any(keyword_item_list):
+                                if keyword_item_list and any(
+                                    keyword_item_list
+                                ):
                                     self.check_parameter_values(
                                         c_func,
                                         (
@@ -256,7 +278,8 @@ class Quark:
 
         return state
 
-    def check_parameter_values(self, source_str, pattern_list, keyword_item_list):
+    @staticmethod
+    def check_parameter_values(source_str, pattern_list, keyword_item_list):
         for pattern, keyword_item in zip(pattern_list, keyword_item_list):
             if keyword_item is None:
                 continue
@@ -283,6 +306,51 @@ class Quark:
 
         return True
 
+    def find_api_usage(self, class_name, method_name, descriptor_name):
+        method_list = []
+
+        # Source method
+        source_method = self.apkinfo.find_method(
+            class_name, method_name, descriptor_name
+        )
+        if source_method:
+            return [source_method]
+
+        # Potential Method
+        potential_method_list = [
+            method
+            for method in self.apkinfo.all_methods
+            if method.name == method_name
+            and method.descriptor == descriptor_name
+        ]
+
+        potential_method_list = [
+            method
+            for method in potential_method_list
+            if not next(self.apkinfo.get_method_bytecode(method), None)
+        ]
+
+        # Check if each method's class is a subclass of the given class
+        for method in potential_method_list:
+            current_class_set = {method.class_name}
+
+            while not current_class_set.intersection(
+                {class_name, "Ljava/lang/Object;"}
+            ):
+                next_class_set = set()
+                for clazz in current_class_set:
+                    next_class_set.update(
+                        self.apkinfo.superclass_relationships[clazz]
+                    )
+
+                current_class_set = next_class_set
+
+            current_class_set.discard("Ljava/lang/Object;")
+            if current_class_set:
+                method_list.append(method)
+
+        return method_list
+
     def run(self, rule_obj):
         """
         Run the five levels check to get the y_score.
@@ -311,79 +379,92 @@ class Quark:
         api_2_class_name = rule_obj.api[1]["class"]
         api_2_descriptor = rule_obj.api[1]["descriptor"]
 
-        first_api = self.apkinfo.find_method(
+        first_api_list = self.find_api_usage(
             api_1_class_name, api_1_method_name, api_1_descriptor
         )
-        second_api = self.apkinfo.find_method(
+        second_api_list = self.find_api_usage(
             api_2_class_name, api_2_method_name, api_2_descriptor
         )
 
-        if first_api is None and second_api is None:
+        if not first_api_list and not second_api_list:
             # Exit if the level 2 stage check fails.
             return
 
         else:
             rule_obj.check_item[1] = True
 
-        if first_api is not None:
-            first_api = self.apkinfo.find_method(
-                api_1_class_name, api_1_method_name, api_1_descriptor
-            )
-            self.quark_analysis.level_2_result.append(first_api)
-        if second_api is not None:
-            second_api = self.apkinfo.find_method(
-                api_2_class_name, api_2_method_name, api_2_descriptor
-            )
-            self.quark_analysis.level_2_result.append(second_api)
+        if first_api_list:
+            self.quark_analysis.level_2_result.append(first_api_list[0])
+        if second_api_list:
+            self.quark_analysis.level_2_result.append(second_api_list[0])
+
         # Level 3: Both Native API Check
-        if first_api is None or second_api is None:
+        if not (first_api_list and second_api_list):
             # Exit if the level 3 stage check fails.
             return
 
-        self.quark_analysis.first_api = first_api
-        self.quark_analysis.second_api = second_api
+        self.quark_analysis.first_api = first_api_list[0]
+        self.quark_analysis.second_api = second_api_list[0]
         rule_obj.check_item[2] = True
 
+        self.quark_analysis.level_3_result = [set(), set()]
+
         # Level 4: Sequence Check
-        # Looking for the first layer of the upper function
-        first_api_xref_from = self.apkinfo.upperfunc(first_api)
-        second_api_xref_from = self.apkinfo.upperfunc(second_api)
+        for first_api in first_api_list:
+            for second_api in second_api_list:
+                # Looking for the first layer of the upper function
+                first_api_xref_from = self.apkinfo.upperfunc(first_api)
+                second_api_xref_from = self.apkinfo.upperfunc(second_api)
 
-        if not (first_api_xref_from and second_api_xref_from):
-            # Exit if the upper function is not found (for Rizin library).
-            return
-
-        mutual_parent_function_list = self.find_intersection(
-            first_api_xref_from, second_api_xref_from
-        )
-
-        if mutual_parent_function_list is None:
-            # Exit if the level 4 stage check fails.
-            return
-        for parent_function in mutual_parent_function_list:
-            first_wrapper = []
-            second_wrapper = []
-
-            self.find_previous_method(first_api, parent_function, first_wrapper)
-            self.find_previous_method(second_api, parent_function, second_wrapper)
-
-            if self.check_sequence(parent_function, first_wrapper, second_wrapper):
-                rule_obj.check_item[3] = True
-                self.quark_analysis.level_4_result.append(parent_function)
-
-                keyword_item_list = (
-                    rule_obj.api[i].get("keyword", None) for i in range(2)
+                self.quark_analysis.level_3_result[0].update(
+                    first_api_xref_from
+                )
+                self.quark_analysis.level_3_result[1].update(
+                    second_api_xref_from
                 )
 
-                # Level 5: Handling The Same Register Check
-                if self.check_parameter(
-                    parent_function,
-                    first_wrapper,
-                    second_wrapper,
-                    keyword_item_list=keyword_item_list,
-                ):
-                    rule_obj.check_item[4] = True
-                    self.quark_analysis.level_5_result.append(parent_function)
+                mutual_parent_function_list = self.find_intersection(
+                    first_api_xref_from, second_api_xref_from
+                )
+
+                if mutual_parent_function_list is None:
+                    # Exit if the level 4 stage check fails.
+                    return
+                for parent_function in mutual_parent_function_list:
+                    first_wrapper = []
+                    second_wrapper = []
+
+                    self.find_previous_method(
+                        first_api, parent_function, first_wrapper
+                    )
+                    self.find_previous_method(
+                        second_api, parent_function, second_wrapper
+                    )
+
+                    if self.check_sequence(
+                        parent_function, first_wrapper, second_wrapper
+                    ):
+                        rule_obj.check_item[3] = True
+                        self.quark_analysis.level_4_result.append(
+                            parent_function
+                        )
+
+                        keyword_item_list = (
+                            rule_obj.api[i].get("keyword", None)
+                            for i in range(2)
+                        )
+
+                        # Level 5: Handling The Same Register Check
+                        if self.check_parameter(
+                            parent_function,
+                            first_wrapper,
+                            second_wrapper,
+                            keyword_item_list=keyword_item_list,
+                        ):
+                            rule_obj.check_item[4] = True
+                            self.quark_analysis.level_5_result.append(
+                                parent_function
+                            )
 
     def get_json_report(self):
         """
@@ -392,7 +473,9 @@ class Quark:
         :return: json report
         """
 
-        w = Weight(self.quark_analysis.score_sum, self.quark_analysis.weight_sum)
+        w = Weight(
+            self.quark_analysis.score_sum, self.quark_analysis.weight_sum
+        )
         warning = w.calculate()
 
         # Filter out color code in threat level
@@ -423,9 +506,7 @@ class Quark:
         score = rule_obj.score
 
         # Assign level 1 examine result
-        permissions = []
-        if rule_obj.check_item[0]:
-            permissions = rule_obj.permission
+        permissions = rule_obj.permission if rule_obj.check_item[0] else []
 
         # Assign level 2 examine result
         api = []
@@ -433,15 +514,19 @@ class Quark:
             for item2 in self.quark_analysis.level_2_result:
                 api.append(
                     {
-                        "class": repr(item2.class_name),
-                        "method": repr(item2.name),
+                        "class": str(item2.class_name),
+                        "method": str(item2.name),
+                        "descriptor": str(item2.descriptor),
                     }
                 )
 
-        # Assign level 3 examine result
+        # Assign level 3 - list methods calling to the APIs
         combination = []
         if rule_obj.check_item[2]:
-            combination = rule_obj.api
+            for method_list in self.quark_analysis.level_3_result:
+                combination.append(
+                    [method.full_name for method in method_list]
+                )
 
         # Assign level 4 - 5 examine result if exist
         sequnce_show_up = []
@@ -454,7 +539,9 @@ class Quark:
                     {
                         repr(
                             item4.full_name
-                        ): self.quark_analysis.parent_wrapper_mapping[item4.full_name]
+                        ): self.quark_analysis.parent_wrapper_mapping[
+                            item4.full_name
+                        ]
                     }
                 )
 
@@ -612,20 +699,20 @@ class Quark:
                 print(f"\t\t {api.full_name}")
         if rule_obj.check_item[2]:
             colorful_report("3.Native API Combination")
-            print(
-                f"\t\t {rule_obj.api[0]['class']} "
-                f"{rule_obj.api[0]['method']} "
-                f"{rule_obj.api[0]['descriptor']}",
-            )
-            print(
-                f"\t\t {rule_obj.api[1]['class']} "
-                f"{rule_obj.api[1]['method']} "
-                f"{rule_obj.api[1]['descriptor']}",
-            )
+            for numbered_api, method_list in zip(
+                ("First API", "Second API"), self.quark_analysis.level_3_result
+            ):
+                print(f"\t\t {numbered_api} show up in:")
+                if method_list:
+                    for comb_method in method_list:
+                        print(f"\t\t {comb_method.full_name}")
+                else:
+                    print("\t\t None")
+
         if rule_obj.check_item[3]:
 
             colorful_report("4.Native API Sequence")
-            print(f"\t\t Sequence show up in:")
+            print("\t\t Sequence show up in:")
             for seq_method in self.quark_analysis.level_4_result:
                 print(f"\t\t {seq_method.full_name}")
         if rule_obj.check_item[4]:
@@ -634,10 +721,12 @@ class Quark:
             for seq_operation in self.quark_analysis.level_5_result:
                 print(f"\t\t {seq_operation.full_name}")
 
-    def show_call_graph(self):
+    def show_call_graph(self, output_format=None):
         print_info("Creating Call Graph...")
-        for call_graph_analysis in self.quark_analysis.call_graph_analysis_list:
-            call_graph(call_graph_analysis)
+        for (
+            call_graph_analysis
+        ) in self.quark_analysis.call_graph_analysis_list:
+            call_graph(call_graph_analysis, output_format)
         print_success("Call Graph Completed")
 
     def show_rule_classification(self):
