@@ -9,18 +9,22 @@
 import logging
 from datetime import datetime
 
+from quark import config
 from quark.core.struct.registerobject import RegisterObject
 from quark.core.struct.tableobject import TableObject
 
 MAX_REG_COUNT = 40
-TIMESTAMPS = datetime.now().strftime("%Y-%m-%d")
-LOG_FILENAME = f"{TIMESTAMPS}.quark.log"
 log = logging.getLogger(__name__)
 log.setLevel(logging.DEBUG)
-handler = logging.FileHandler(LOG_FILENAME, mode="w")
-format_str = "%(asctime)s %(levelname)s [%(lineno)d]: %(message)s"
-handler.setFormatter(logging.Formatter(format_str))
-log.addHandler(handler)
+if config.DEBUG:
+    TIMESTAMPS = datetime.now().strftime("%Y-%m-%d")
+    LOG_FILENAME = f"{TIMESTAMPS}.quark.log"
+    handler = logging.FileHandler(LOG_FILENAME, mode="w")
+    format_str = "%(asctime)s %(levelname)s [%(lineno)d]: %(message)s"
+    handler.setFormatter(logging.Formatter(format_str))
+    log.addHandler(handler)
+else:
+    log.disabled = True
 
 
 def logger(func):
@@ -648,7 +652,7 @@ class PyEval:
         method_name, descriptor = signature[:index], signature[index:]
 
         class_pool = (
-            self.apkinfo.class_hierarchy[instance_type]
+            self.apkinfo.superclass_relationships[instance_type]
             if skip_self
             else {instance_type}
         )
@@ -663,7 +667,9 @@ class PyEval:
                 if method:
                     return f"{method.class_name}->" f"{method.name}{method.descriptor}"
 
-                next_class_pool.update((self.apkinfo.class_hierarchy[class_name]))
+                next_class_pool.update(
+                    (self.apkinfo.superclass_relationships[class_name])
+                )
                 next_class_pool.difference_update(class_pool)
 
             class_pool = set(next_class_pool)
